@@ -1,5 +1,6 @@
 import random
 import logging
+from collections import Counter
 try:
    from IPython.core.display import display, HTML
 except:
@@ -32,8 +33,10 @@ class Text_Dataset:
         self.index = 1
         self.word_tokenizer_init = False
         self.char_tokenizer_init = False
+        self.ngram_tokenizer_init = False
         self.getitem_init = False
         self.tokenizer_type = None
+        self.special_words=['<unk>', '<pad>', '<sos>', '<eos>', '<wsep>']
         
         req_attrs=['title', 'author', 'language', 'text']
         for ind in range(0,len(text_list)):
@@ -138,29 +141,66 @@ class Text_Dataset:
             sample = text[pos:pos+length]
         return (idx, sample)
 
-    def _word_splitter(self, text):
-        tokens=text.split()
+    def _word_splitter(self, text, word_separator=' '):
+        tokens=text.split(word_separator)
         return tokens
 
-    def init_tokenizer(self, tokenizer='word'):
-        """ Initialize the tokenizer with the text_list.
+    def _every_ngram(self, text, max_len, add_special_words=True):
+        """ Return all ngrams of length 1..max_len from text.
         
-        :param tokenizer: 'word' or 'char'
+        :param text: text to extract ngrams from
+        :param max_len: maximum length of ngrams to extract
+        :param add_special_words: If True, special words ('<unk>' etc.) are added to the ngrams.
+        """
+        ngrams = [text[i:i+j+1] for i in range(len(text)) for j in range(0, min(len(text)-i,max_len))]
+        if add_special_words is True:
+            ngrams = self.special_words + ngrams            
+        return ngrams
+
+    def _weight_ngrams(self, ngrams, max_weight=1e12):
+        """ Weight ngrams by their length. Ngrams of length==1 and special words ('<unk>' etc.) are 
+        weigted by max_weight, since we need the full character set and the special words to be 
+        included in the ngram collection.
+        """
+        eg_dict=Counter(ngrams)
+        return sorted([(''.join(l), max_weight if len(l)==1 or l in self.special_words else len(l)*eg_dict[l]) for l in eg_dict.keys()], key=lambda x: x[1], reverse=True) 
+
+    def init_tokenizer(self, tokenizer='word', max_ngrams=5, word_separator=None, max_tokens=5000):
+        """ Initialize the tokenizer with the text_list.
+
+        The character tokenizer simply splits any language text into glyphs. 
+        
+        A word tokenizer works for languages that have the concept of word separators (as in English, 
+        ' ' space). Word tokenizers are English-centric algorithms, since many other languages either
+        can compose words into larger words with no separators, words being more similar to English
+        syllables in that case, or have no word separators at all (as in Chinese), or have syllable
+        separators (as in Tibetan, which has also a second kind of separators for sentence-fragments). For multi-language applications, use ngram tokenizers or character
+        tokenizers.
+        
+        The ngram tokenizer can either first split text into words using a word_speparator 
+        and then extracting all possible ngrams, or it can extract all possible ngrams directly 
+        without any word-splitting, which is the default behavior with word_separator=None since it 
+        works for all languages classes.
+
+        For ngrams, the tokenizer can extract ngrams of length 1..max_ngrams, and selects the top most
+        used ngrams with upper limit max_tokens. The optimum for max_ngrams is usually around 3-6, and
+        max_tokens should be significantly higher than the number of unique glyphs in the text_list.
+        Using word_separator=None is usually significantly better than using a word_separator for ngrams.
+
+        :param tokenizer: 'word', 'char', or 'ngram'
+        :param max_ngrams: (ngram only) maximum n-gram length
+        :param word_separator: (word, ngram) character used to separate words, default None, which amounts to ' ' (space) for word and no word-splitting for ngram.
+        :param max_tokens: (ngram only) maximum number of tokens to use
         """
         if tokenizer == 'word':
             self.tokenizer_type = 'word'
             self.w2i = {}
             self.i2w = {}
-            self.w2i['<unk>'] = 0
-            self.i2w[0] = '<unk>'
-            self.w2i['<pad>'] = 1
-            self.i2w[1] = '<pad>'
-            self.w2i['<eos>'] = 2
-            self.i2w[2] = '<eos>'
-            self.w2i['<sos>'] = 3
-            self.i2w[3] = '<sos>'
+            for ind, tok in enumerate(self.special_words):
+                self.w2i[tok] = ind
+                self.i2w[ind] = tok
             for text in self.text_list:
-                tokens = self._word_splitter(text['text'])
+                tokens = self._word_splitter(text['text'], word_separator=word_separator)
                 for token in tokens:
                     if token not in self.w2i:
                         self.w2i[token] = len(self.w2i)
@@ -187,9 +227,42 @@ class Text_Dataset:
                         self.c2i[c] = ind
                         self.i2c[ind] = c
             self.char_tokenizer_init=True
+        elif tokenizer == 'ngram':
+            self.t2i={}
+            self.i2t={}
+            self.tokenizer_type = 'ngram'
+            self.word_separator = word_separator
+            self.max_ngrams = max_ngrams
+            self.log.info(f"Extracting ngrams of length 1..{max_ngrams} from text_list, selecting {max_tokens} most used ngrams.")
+            corpus=""
+            for text in self.text_list:
+                corpus+=text['text']   # TODO: This generates ngrams across text-bordes, should be changed at some point.
+            if word_separator is not None:
+                self.word_list=corpus.split(word_separator)
+            else:
+                self.word_list=None
+            if self.word_list is None:
+                ngrams=self._every_ngram(corpus,max_len=max_ngrams)
+                self.ngrams_list = self._weight_ngrams(ngrams)
+            else:
+                ngrams=[]
+                for word in self.word_list:
+                    ngrams += self._every_ngram(word, max_len=max_ngrams)
+                self.ngrams_list = self._weight_ngrams(ngrams)
+            if max_tokens is not None:
+                if len(self.ngrams_list)>max_tokens:
+                    self.ngrams_list=self.ngrams_list[:max_tokens]
+            self.t2i={t[1][0]: t[0] for t in enumerate(self.ngrams_list)}
+            self.ngram_tokenizer_init=True
         else:
             self.log.error(f"Unknown tokenizer {tokenizer}")
             raise ValueError(f"Unknown tokenizer {tokenizer}")
+            return
+        self.log.info("Encoding text corpora as ngrams.")
+        for text in self.text_list:
+            self.log.info(f"Encoding text {text['title']}...")
+            text['text_encoded']=self.encode(text['text'])
+        self.log.info("Encoding text corpora as ngrams done.")
 
     def tokenize(self, text):
         """ Tokenize a text.
@@ -199,12 +272,48 @@ class Text_Dataset:
         tokens = []
         if self.tokenizer_type == 'word':
             if self.word_tokenizer_init is False:
-                self.init_tokenizer(tokenizer)
+                self.init_tokenizer(tokenizer='word')
             tokens = self._word_splitter(text)
         elif self.tokenizer_type == 'char':
             if self.char_tokenizer_init is False:
-                self.init_tokenizer(tokenizer)
+                self.init_tokenizer(tokenizer='char')
             tokens = list(text)
+        elif self.tokenizer_type == 'ngram':
+            if self.ngram_tokenizer_init is False:
+                self.init_tokenizer(tokenizer='ngram')
+            if self.word_separator is not None:
+                wrd_list=text.split(self.word_separator)
+                for word in wrd_list:
+                    while len(word)>0:
+                        mx = min(self.max_ngrams,len(word))            
+                        for si in range(mx,0,-1):
+                            tk=word[:si]
+                            if tk in self.t2i:
+                                ind = self.t2i[tk]
+                                word = word[si:]
+                                tokens.append(ind)
+                                break
+                        if len(word)>0:
+                            if word[0] not in self.t2i:
+                                tokens.append('<unk>')  # unknown encounter
+                                word=word[1:]  # throw away one char
+                    tokens.append('<wsep>')  # word separator
+                if len(wrd_list)>1:
+                    tokens=tokens[:-1]  # remove last seperator
+            else:
+                while len(text)>0:
+                    mx = min(self.max_ngrams,len(text))            
+                    for si in range(mx,0,-1):
+                        tk=text[:si]
+                        if tk in self.t2i:
+                            ind = self.t2i[tk]
+                            text = text[si:]
+                            tokens.append(ind)
+                            break
+                    if len(text)>0:
+                        if text[0] not in self.t2i:
+                            tokens.append('<unk>')
+                            text=text[1:]
         else:
             self.log.error(f"Unknown tokenizer {self.tokenizer_type}")
             raise ValueError(f"Unknown tokenizer {self.tokenizer_type}")
@@ -220,15 +329,18 @@ class Text_Dataset:
             encoded = [self.w2i[token] if token in self.w2i else self.w2i['<unk>'] for token in tokens]
         elif self.tokenizer_type == 'char':
             encoded = [self.c2i[token] if token in self.c2i else self.c2i['␚'] for token in tokens]
+        elif self.tokenizer_type == 'ngram':
+            encoded = tokens
         else:
             self.log.error(f"Unknown tokenizer {self.tokenizer_type}")
             raise ValueError(f"Unknown tokenizer {self.tokenizer_type}")
         return encoded
 
-    def decode(self, encoded):
+    def decode(self, encoded, mark_separator=False):
         """ Decode a list of encoded tokens.
         
         :param encoded: list of encoded tokens
+        :param mark_separator: (ngram only) if True, add a separator between tokens for debug
         :return: text """
         if self.tokenizer_type == 'word':
             decoded = [self.i2w[token]+'' if token in self.i2w else '<unk>' for token in encoded]
@@ -236,11 +348,37 @@ class Text_Dataset:
         elif self.tokenizer_type == 'char':
             decoded = [self.i2c[token] if token in self.i2c else '␚' for token in encoded]
             decoded_text = ''.join(decoded)
+        elif self.tokenizer_type == 'ngram':
+            dec=""
+            for ind in encoded:
+                if ind=='<wsep>':  # Separator
+                    dec+=self.word_separator
+                elif ind=='<unk>':  # Unknown token
+                    dec+="<unk>"
+                else:
+                    dec+=self.ngrams_list[ind][0]
+                    if mark_separator is True:
+                        dec+='_'
+            decoded_text = dec
         else:
             self.log.error(f"Unknown tokenizer {self.tokenizer_type}")
             raise ValueError(f"Unknown tokenizer {self.tokenizer_type}")
         return decoded_text
 
+    def get_unique_token_count(self):
+        """ Get the number of unique tokens.
+        
+        :return: number of unique tokens """
+        if self.tokenizer_type == 'word':
+            return len(self.w2i)
+        elif self.tokenizer_type == 'char':
+            return len(self.c2i)
+        elif self.tokenizer_type == 'ngram':
+            return len(self.t2i)
+        else:
+            self.log.error(f"Unknown tokenizer {self.tokenizer_type}")
+            raise ValueError(f"Unknown tokenizer {self.tokenizer_type}")
+
     def get_random_char_tokenized_sample_pair(self, length):
         """ Get a random tokenized sample of the dataset.
         
@@ -253,7 +391,7 @@ class Text_Dataset:
         y = e_sample[1:]
         return X, y
 
-    def init_getitem(self, sample_type='chargen', sample_length=80, content_stepping=10):
+    def init_getitem(self, sample_type='text', sample_length=80, content_stepping=10):
         """ Initialize the __getitem__ and __len__ methods.
 
         This method needs to be called before using len() or index-access of the dataset.
@@ -267,36 +405,46 @@ class Text_Dataset:
             tl = [{'author':'nobody', 'title':'some title', 'language':'english', 'text':'some text'},
                   {'author':'nobody', 'title':'some title 2', 'language':'english', 'text':'some more text'}]
             td = Text_Dataset(tl)
-            td.init_getitem(sample_type='chargen', sample_length=4, content_stepping=2)
+            td.init_getitem(sample_type='text', sample_length=4, content_stepping=2)
             print(len(td))
             print(td[0])
             # Output: 12 and ('some', 'ome ')
 
-        :param sample_type: 'chargen': generate a pair of text or length sample_length, shifted by one letter, or 'chargen_encoded', same but encoded. 'chargen_single_encoded' just returns a single encoded string X.
-        :param sample_length: length of a sample
-        :param content_stepping: number of characters to skip between each sample
-        :return: on 'chargen[_encoded]': X, y [encoded] strings of sample_length, y shifted by one letter; for 'chargin_single_encoded' just encoded X.
+        The method of tokenization (char, word, ngram) for 'encoded' sample_type is determined by 
+        the tokenizer_type in call to init_tokenizer().
+
+        :param sample_type: 'text' (text-string of length sample_length), 'encoded' (encoded sample_length tokens) 
+        :param sample_length: length of a sample (either character count (type text) or token count (type encoded))
+        :param content_stepping: number of characters/tokens to skip between each sample
         """
 
         self.getitem_sample_type = sample_type
         self.getitem_sample_length = sample_length
         self.getitem_content_stepping = content_stepping
-        leng=0
-        rec=0
-        if sample_type=='chargen' or sample_type=='chargen_encoded' or sample_type=='chargen_single_encoded':
-            for ind in range(0, len(self.text_list)):
-                len_text = len(self.text_list[ind]['text'])
-                rec_text = (len_text-content_stepping+1)//content_stepping + 1
-                self.text_list[ind]['records']=rec_text
-                leng += len_text
-                rec += rec_text
-            self.getitem_length = leng
-            self.getitem_records = rec
-            self.getitem_init = True
-        else:
+        len_tot=0
+        rec_tot=0
+        if sample_type not in ['text','encoded']:
+            self.log.error(f"Unknown sample type {sample_type}")
             self.getitem_length = 0
             self.getitem_records = 0
-            print(f"init_getitem: unknown sample_type {sample_type}")
+            return None
+        for ind in range(0, len(self.text_list)):
+            if sample_type == 'text':
+                leni = len(self.text_list[ind]['text'])
+            elif sample_type == 'encoded':
+                leni = len(self.text_list[ind]['text_encoded'])
+            else:
+                self.log.error(f"Unknown sample type {sample_type}")
+                self.getitem_length = 0
+                self.getitem_records = 0
+                return None
+            recs = (leni-content_stepping+1)//content_stepping + 1
+            self.text_list[ind]['records']=recs
+            len_tot += leni
+            rec_tot += recs
+        self.getitem_length = len_tot
+        self.getitem_records = rec_tot
+        self.getitem_init = True
 
     def __len__(self):
         """ Get the length of the dataset.
@@ -306,11 +454,21 @@ class Text_Dataset:
         :return: length of the dataset (mode dependent) 
         """
         if self.getitem_init is False:
-            print("init_getitem must be called before __len__")
+            self.log.error("init_getitem must be called before __len__")
             return None
         return self.getitem_records
 
-    def _getitem_chargen(self, index):
+    def __getitem__(self, index):
+        """ Get a sample from the dataset.
+
+        Format of the returned sample depends on :ref:`~Text_Dataset.Text_Dataset.init_getitem`.
+
+        :param index: index of the sample
+        :return:
+        """
+        if self.getitem_init is False:
+            print("init_getitem must be called before __getitem__")
+            raise ValueError("init_getitem must be called before __getitem__")
         if index<0:
             if index<(-self.getitem_records):
                 raise IndexError(f"index {index} out of range")
@@ -324,53 +482,39 @@ class Text_Dataset:
             if cur_rec+rec > index:
                 rel_rec = index - cur_rec
                 pos = rel_rec*self.getitem_content_stepping
-                if self.getitem_sample_type=='chargen_encoded' or self.getitem_sample_type=='chargen':
+                if self.getitem_sample_type == 'text':
                     sample = text['text'][pos:pos+self.getitem_sample_length+1]
                     while len(sample) < self.getitem_sample_length+1:
-                        sample += ' '
-                    X_text = sample[:-1]
-                    y_text = sample[1:]
-                    return X_text, y_text
-                elif self.getitem_sample_type=='chargen_single_encoded':
-                    sample = text['text'][pos:pos+self.getitem_sample_length]
-                    while len(sample) < self.getitem_sample_length:
-                        sample += ' '
-                    X_text = sample
-                    return X_text
+                        if self.tokenizer_type=='char':
+                            sample += '␠'  # pad with ␠ character
+                        elif self.tokenizer_type=='ngram':
+                            sample += '<unk>'
+                        elif self.tokenizer_type=='word':
+                            sample += '<unk>'
+                        else:
+                            self.log.error(f"Unknown tokenizer {self.tokenizer_type}")
+                            raise ValueError(f"Unknown tokenizer {self.tokenizer_type}")
+                    return sample
+                elif self.getitem_sample_type == 'encoded':
+                    sample = text['text_encoded'][pos:pos+self.getitem_sample_length+1]
+                    while len(sample) < self.getitem_sample_length+1:
+                        if self.tokenizer_type=='char':
+                            sample += self.c2i['␠']  # pad with ␠ character
+                        elif self.tokenizer_type=='ngram':
+                            sample += self.t2i['<unk>']
+                        elif self.tokenizer_type=='word':
+                            sample += self.w2i['<unk>']
+                        else:
+                            self.log.error(f"Unknown tokenizer {self.tokenizer_type}")
+                            raise ValueError(f"Unknown tokenizer {self.tokenizer_type}")
+                    return sample
                 else:
-                    print(f"_getitem_chargen: unknown sample_type {self.getitem_sample_type}")
+                    self.log.error(f"Unknown sample type {self.getitem_sample_type}")
                     return None
-            cur_rec += rec
-            
+            cur_rec += rec        
         print("Internal error in __getitem__")
         raise ValueError("Internal error in __getitem__")
 
-    def __getitem__(self, index):
-        """ Get a sample from the dataset.
-
-        Format of the returned sample depends on :ref:`~Text_Dataset.Text_Dataset.init_getitem`.
-
-        :param index: index of the sample
-        :return:
-        """
-        if self.getitem_init is False:
-            print("init_getitem must be called before __getitem__")
-            raise ValueError("init_getitem must be called before __getitem__")
-        if self.getitem_sample_type == 'chargen':
-            return self._getitem_chargen(index)
-        elif self.getitem_sample_type == 'chargen_encoded':
-            X, y = self._getitem_chargen(index)
-            X = self.encode(X)
-            y = self.encode(y)
-            return X, y
-        elif self.getitem_sample_type == 'chargen_single_encoded':
-            X = self._getitem_chargen(index)
-            X = self.encode(X)
-            return X
-        else:
-            self.log.error(f"Unknown getitem sample_type {self.getitem_sample_type}")
-            raise ValueError(f"Unknown getitem sample_type {self.getitem_sample_type}")
-
     def _display_colored_html(self, textlist, dark_mode=False, display_ref_anchor=True, pre='', post=''):
         """ Internal function to display text and citation references in HTML. """
         bgcolorsWht = ['#d4e6e1', '#d8daef', '#ebdef0', '#eadbd8', '#e2d7d5', '#edebd0',
