@@ -1,6 +1,13 @@
 import tensorflow as tf
-import tensorflow.keras as keras
-from tensorflow.keras import layers
+try:
+    # the endless shuffle of keras modules
+    import tensorflow.keras as keras
+    from tensorflow.keras import layers
+    print("Using TF-Keras version:", keras.__version__)
+except ImportError:
+    import keras
+    import keras.layers as layers
+    print("Using Keras version:", keras.__version__)
 import numpy as np
 import math
 
